//! Reference Solar Spectrum
//!
//! This is the 2000 ASTM Standard Extraterrestrial Spectrum Reference E-490-00:
//! <https://www.nrel.gov/grid/solar-resource/spectra-astm-e490.html>
//!

use std::f64::consts::PI;

use crate::{
    constants::{AU_KM, C_M_PER_S, SUN_DIAMETER, SUN_TEMP},
    prelude::black_body_flux,
};

/// Return the Solar flux at the provide wavelength using the 2000 ASTM Reference Spectrum.
///
/// This does linear interpolation between the two closest data points.
///
/// Data is available from 119.5 nm to 1mm.
///
/// If query is outside of the range of data, then [`None`] is returned.
///
/// Returned values are units Janskys / steradian.
///
/// # Arguments
///
/// * `dist` - Distance from the Sun in au.
/// * `wavelength` - wavelength in nm.
pub fn solar_flux(dist: f64, wavelength: f64) -> Option<f64> {
    let wavelength = wavelength / 1000.0;
    let start_idx = match E490_DATA.binary_search_by(|probe| probe[0].total_cmp(&wavelength)) {
        Ok(c) => c,
        Err(c) => {
            if c >= E490_DATA.len() - 1 || c == 0 {
                return None;
            };
            c - 1
        }
    };

    let low = E490_DATA[start_idx];
    let high = E490_DATA[start_idx + 1];

    let w_frac = (wavelength - low[0]) / (high[0] - low[0]);
    let val = w_frac * (high[1] - low[1]) + low[1];

    // convert from Watts / m^2 / micron to Jy / Steradian
    Some(val * 1e20 * wavelength.powi(2) / C_M_PER_S / dist.powi(2))
}

/// Calculate the solar flux at the object assuming the sun is a black body disk
/// emitting at the specified wavelength.
///
/// This is an approximation which breaks down when the object gets within a few solar
/// radii.
///
/// This returns the flux at the object in Jy / steradian per unit frequency.
///
/// # Arguments
///
/// * `sun2obj` - Vector from the sun to the object in AU.
/// * `wavelength` - Central wavelength of light in nm.
///
#[inline(always)]
pub fn solar_flux_black_body(dist: f64, wavelength: f64) -> f64 {
    const SUN_R_AU: f64 = SUN_DIAMETER / AU_KM / 2.0;

    // blackbody radiation is Janskys / steradian
    let solar_flux = black_body_flux(SUN_TEMP, wavelength);

    // solar flux at the radius of where the object is, this treats the sun as a flat
    // disk facing the object. This is an approximation which breaks down when the object
    // gets within a few solar radii.
    // Jy / steradian per unit freq
    PI * solar_flux * (SUN_R_AU / dist).powi(2)
}

/// The 2000 ASTM Standard Extraterrestrial Spectrum Reference E-490-00:
/// <https://www.nrel.gov/grid/solar-resource/spectra-astm-e490.html>
///
/// First column is wavelength in um, second column is Watts / m^2 / micron at 1 AU.
const E490_DATA: &[[f64; 2]] = &[
    [0.1195, 0.0619],
    [0.1205, 0.5614],
    [0.1215, 4.901],
    [0.1225, 1.184],
    [0.1235, 0.0477],
    [0.1245, 0.0343],
    [0.1255, 0.0288],
    [0.1265, 0.0352],
    [0.1275, 0.0213],
    [0.1285, 0.0173],
    [0.1295, 0.0399],
    [0.1305, 0.1206],
    [0.1315, 0.0398],
    [0.1325, 0.0413],
    [0.1335, 0.168],
    [0.1345, 0.0457],
    [0.1355, 0.038],
    [0.1365, 0.0309],
    [0.1375, 0.0292],
    [0.1385, 0.0397],
    [0.1395, 0.0756],
    [0.1405, 0.0608],
    [0.1415, 0.0421],
    [0.1425, 0.0468],
    [0.1435, 0.0511],
    [0.1445, 0.0509],
    [0.1455, 0.0554],
    [0.1465, 0.0709],
    [0.1475, 0.0849],
    [0.1485, 0.082],
    [0.1495, 0.0796],
    [0.1505, 0.087],
    [0.1515, 0.0927],
    [0.1525, 0.1163],
    [0.1535, 0.1299],
    [0.1545, 0.2059],
    [0.1555, 0.2144],
    [0.1565, 0.1847],
    [0.1575, 0.1717],
    [0.1585, 0.1675],
    [0.1595, 0.1754],
    [0.1605, 0.1934],
    [0.1615, 0.2228],
    [0.1625, 0.2519],
    [0.1635, 0.2841],
    [0.1645, 0.2973],
    [0.1655, 0.4302],
    [0.1665, 0.3989],
    [0.1675, 0.3875],
    [0.1685, 0.4556],
    [0.1695, 0.5877],
    [0.1705, 0.6616],
    [0.1715, 0.688],
    [0.1725, 0.7252],
    [0.1735, 0.7645],
    [0.1745, 0.9067],
    [0.1755, 1.079],
    [0.1765, 1.22],
    [0.1775, 1.403],
    [0.1785, 1.538],
    [0.1795, 1.576],
    [0.1805, 1.831],
    [0.1815, 2.233],
    [0.1825, 2.243],
    [0.1835, 2.244],
    [0.1845, 2.066],
    [0.1855, 2.311],
    [0.1865, 2.7],
    [0.1875, 3.009],
    [0.1885, 3.291],
    [0.1895, 3.569],
    [0.1905, 3.764],
    [0.1915, 4.165],
    [0.1925, 4.113],
    [0.1935, 3.808],
    [0.1945, 5.21],
    [0.1955, 5.427],
    [0.1965, 6.008],
    [0.1975, 6.191],
    [0.1985, 6.187],
    [0.1995, 6.664],
    [0.2005, 7.326],
    [0.2015, 8.023],
    [0.2025, 8.261],
    [0.2035, 9.217],
    [0.2045, 10.25],
    [0.2055, 10.54],
    [0.2065, 11.08],
    [0.2075, 12.65],
    [0.2085, 15.05],
    [0.2095, 21.38],
    [0.2105, 27.92],
    [0.2115, 33.54],
    [0.2125, 31.3],
    [0.2135, 33.15],
    [0.2145, 40.03],
    [0.2155, 36.15],
    [0.2165, 32.27],
    [0.2175, 35.29],
    [0.2185, 44.37],
    [0.2195, 46.92],
    [0.2205, 47.33],
    [0.2215, 39.58],
    [0.2225, 49.65],
    [0.2235, 63.01],
    [0.2245, 58.97],
    [0.2255, 52.29],
    [0.2265, 39.4],
    [0.2275, 39.92],
    [0.2285, 51.95],
    [0.2295, 47.71],
    [0.2305, 52.12],
    [0.2315, 50.97],
    [0.2325, 53.26],
    [0.2335, 44.74],
    [0.2345, 38.97],
    [0.2355, 51.42],
    [0.2365, 48.59],
    [0.2375, 48.44],
    [0.2385, 41.96],
    [0.2395, 44.12],
    [0.2405, 39.56],
    [0.2415, 51.48],
    [0.2425, 70.6],
    [0.2435, 66.53],
    [0.2445, 60.97],
    [0.2455, 49.39],
    [0.2465, 50.4],
    [0.2475, 55.5],
    [0.2485, 45.65],
    [0.2495, 56.38],
    [0.2505, 60.1],
    [0.2515, 46.01],
    [0.2525, 41.55],
    [0.2535, 51.55],
    [0.2545, 59.57],
    [0.2555, 79.3],
    [0.2565, 101.8],
    [0.2575, 125.4],
    [0.2585, 125.1],
    [0.2595, 104.0],
    [0.2605, 85.51],
    [0.2615, 89.8],
    [0.2625, 103.6],
    [0.2635, 165.8],
    [0.2645, 249.7],
    [0.2655, 252.7],
    [0.2665, 249.4],
    [0.2675, 250.8],
    [0.2685, 243.8],
    [0.2695, 238.9],
    [0.2705, 267.3],
    [0.2715, 224.4],
    [0.2725, 197.4],
    [0.2735, 196.5],
    [0.2745, 132.6],
    [0.2755, 175.1],
    [0.2765, 242.8],
    [0.2775, 233.8],
    [0.2785, 159.3],
    [0.2795, 85.55],
    [0.2805, 94.63],
    [0.2815, 208.3],
    [0.2825, 294.1],
    [0.2835, 313.5],
    [0.2845, 235.3],
    [0.2855, 163.1],
    [0.2865, 322.7],
    [0.2875, 336.3],
    [0.2885, 322.2],
    [0.2895, 472.7],
    [0.2905, 601.3],
    [0.2915, 580.8],
    [0.2925, 521.9],
    [0.2935, 535.5],
    [0.2945, 508.8],
    [0.2955, 553.2],
    [0.2965, 509.6],
    [0.2975, 507.3],
    [0.2985, 465.5],
    [0.2995, 484.0],
    [0.3005, 420.0],
    [0.3015, 455.5],
    [0.3025, 489.0],
    [0.3035, 620.6],
    [0.3045, 602.5],
    [0.3055, 594.8],
    [0.3065, 555.7],
    [0.3075, 615.0],
    [0.3085, 611.4],
    [0.3095, 496.5],
    [0.3105, 622.4],
    [0.3115, 729.2],
    [0.3125, 655.9],
    [0.3135, 699.9],
    [0.3145, 662.9],
    [0.3155, 633.0],
    [0.3165, 633.2],
    [0.3175, 773.9],
    [0.3185, 664.9],
    [0.3195, 710.5],
    [0.3205, 805.1],
    [0.3215, 699.5],
    [0.3225, 688.6],
    [0.3235, 661.3],
    [0.3245, 760.8],
    [0.3255, 875.8],
    [0.3265, 979.5],
    [0.3275, 952.7],
    [0.3285, 917.6],
    [0.3295, 1061.0],
    [0.3305, 1016.0],
    [0.3315, 965.7],
    [0.3325, 954.9],
    [0.3335, 921.6],
    [0.3345, 958.9],
    [0.3355, 943.4],
    [0.3365, 809.5],
    [0.3375, 841.8],
    [0.3385, 921.5],
    [0.3395, 958.1],
    [0.3405, 1007.0],
    [0.3415, 923.8],
    [0.3425, 993.0],
    [0.3435, 950.6],
    [0.3445, 795.7],
    [0.3455, 939.2],
    [0.3465, 926.4],
    [0.3475, 901.7],
    [0.3485, 897.2],
    [0.3495, 889.8],
    [0.3505, 1050.0],
    [0.3515, 979.5],
    [0.3525, 907.9],
    [0.3535, 1033.0],
    [0.3545, 1111.0],
    [0.3555, 1045.0],
    [0.3565, 912.3],
    [0.3575, 796.0],
    [0.3585, 693.6],
    [0.3595, 991.1],
    [0.3605, 970.8],
    [0.3615, 878.1],
    [0.3625, 997.8],
    [0.3635, 996.9],
    [0.3645, 1013.0],
    [0.3655, 1152.0],
    [0.3665, 1233.0],
    [0.3675, 1180.0],
    [0.3685, 1101.0],
    [0.3695, 1226.0],
    [0.3705, 1139.0],
    [0.3715, 1175.0],
    [0.3725, 1054.0],
    [0.3735, 920.2],
    [0.3745, 900.4],
    [0.3755, 1062.0],
    [0.3765, 1085.0],
    [0.3775, 1282.0],
    [0.3785, 1327.0],
    [0.3795, 1066.0],
    [0.3805, 1202.0],
    [0.3815, 1082.0],
    [0.3825, 791.3],
    [0.3835, 684.1],
    [0.3845, 959.7],
    [0.3855, 1008.0],
    [0.3865, 1007.0],
    [0.3875, 1004.0],
    [0.3885, 984.3],
    [0.3895, 1174.0],
    [0.3905, 1247.0],
    [0.3915, 1342.0],
    [0.3925, 1019.0],
    [0.3935, 582.3],
    [0.3945, 1026.0],
    [0.3955, 1314.0],
    [0.3965, 854.5],
    [0.3975, 928.8],
    [0.3985, 1522.0],
    [0.3995, 1663.0],
    [0.4005, 1682.0],
    [0.4015, 1746.0],
    [0.4025, 1759.0],
    [0.4035, 1684.0],
    [0.4045, 1674.0],
    [0.4055, 1667.0],
    [0.4065, 1589.0],
    [0.4075, 1628.0],
    [0.4085, 1735.0],
    [0.4095, 1715.0],
    [0.4105, 1532.0],
    [0.4115, 1817.0],
    [0.4125, 1789.0],
    [0.4135, 1756.0],
    [0.4145, 1737.0],
    [0.4155, 1734.0],
    [0.4165, 1842.0],
    [0.4175, 1665.0],
    [0.4185, 1684.0],
    [0.4195, 1701.0],
    [0.4205, 1757.0],
    [0.4215, 1797.0],
    [0.4225, 1582.0],
    [0.4235, 1711.0],
    [0.4245, 1767.0],
    [0.4255, 1695.0],
    [0.4265, 1698.0],
    [0.4275, 1569.0],
    [0.4285, 1587.0],
    [0.4295, 1475.0],
    [0.4305, 1135.0],
    [0.4315, 1686.0],
    [0.4325, 1646.0],
    [0.4335, 1731.0],
    [0.4345, 1670.0],
    [0.4355, 1723.0],
    [0.4365, 1929.0],
    [0.4375, 1806.0],
    [0.4385, 1567.0],
    [0.4395, 1825.0],
    [0.4405, 1713.0],
    [0.4415, 1931.0],
    [0.4425, 1980.0],
    [0.4435, 1909.0],
    [0.4445, 1973.0],
    [0.4455, 1821.0],
    [0.4465, 1891.0],
    [0.4475, 2077.0],
    [0.4485, 1973.0],
    [0.4495, 2027.0],
    [0.4505, 2144.0],
    [0.4515, 2109.0],
    [0.4525, 1941.0],
    [0.4535, 1970.0],
    [0.4545, 1979.0],
    [0.4555, 2034.0],
    [0.4565, 2077.0],
    [0.4575, 2100.0],
    [0.4585, 1971.0],
    [0.4595, 2009.0],
    [0.4605, 2040.0],
    [0.4615, 2055.0],
    [0.4625, 2104.0],
    [0.4635, 2040.0],
    [0.4645, 1976.0],
    [0.4655, 2042.0],
    [0.4665, 1921.0],
    [0.4675, 2015.0],
    [0.4685, 1994.0],
    [0.4695, 1990.0],
    [0.4705, 1877.0],
    [0.4715, 2018.0],
    [0.4725, 2041.0],
    [0.4735, 1991.0],
    [0.4745, 2051.0],
    [0.4755, 2016.0],
    [0.4765, 1956.0],
    [0.4775, 2075.0],
    [0.4785, 2009.0],
    [0.4795, 2076.0],
    [0.4805, 2035.0],
    [0.4815, 2090.0],
    [0.4825, 2023.0],
    [0.4835, 2019.0],
    [0.4845, 1969.0],
    [0.4855, 1830.0],
    [0.4865, 1625.0],
    [0.4875, 1830.0],
    [0.4885, 1914.0],
    [0.4895, 1960.0],
    [0.4905, 2007.0],
    [0.4915, 1896.0],
    [0.4925, 1896.0],
    [0.4935, 1888.0],
    [0.4945, 2058.0],
    [0.4955, 1926.0],
    [0.4965, 2017.0],
    [0.4975, 2018.0],
    [0.4985, 1866.0],
    [0.4995, 1970.0],
    [0.5005, 1857.0],
    [0.5015, 1812.0],
    [0.5025, 1894.0],
    [0.5035, 1934.0],
    [0.5045, 1869.0],
    [0.5055, 1993.0],
    [0.5065, 1961.0],
    [0.5075, 1906.0],
    [0.5085, 1919.0],
    [0.5095, 1916.0],
    [0.5105, 1947.0],
    [0.5115, 1997.0],
    [0.5125, 1867.0],
    [0.5135, 1861.0],
    [0.5145, 1874.0],
    [0.5155, 1900.0],
    [0.5165, 1669.0],
    [0.5175, 1726.0],
    [0.5185, 1654.0],
    [0.5195, 1828.0],
    [0.5205, 1831.0],
    [0.5215, 1906.0],
    [0.5225, 1823.0],
    [0.52350, 1894.0],
    [0.5245, 1958.0],
    [0.5255, 1930.0],
    [0.5265, 1674.0],
    [0.5275, 1828.0],
    [0.5285, 1897.0],
    [0.5295, 1918.0],
    [0.5305, 1952.0],
    [0.5315, 1963.0],
    [0.5325, 1770.0],
    [0.5335, 1923.0],
    [0.5345, 1858.0],
    [0.5355, 1990.0],
    [0.5365, 1871.0],
    [0.5375, 1882.0],
    [0.5385, 1904.0],
    [0.5395, 1832.0],
    [0.5405, 1769.0],
    [0.5415, 1881.0],
    [0.5425, 1825.0],
    [0.5435, 1879.0],
    [0.5445, 1879.0],
    [0.5455, 1901.0],
    [0.5465, 1879.0],
    [0.5475, 1833.0],
    [0.5485, 1863.0],
    [0.5495, 1895.0],
    [0.5505, 1862.0],
    [0.5515, 1871.0],
    [0.5525, 1846.0],
    [0.5535, 1882.0],
    [0.5545, 1898.0],
    [0.5555, 1897.0],
    [0.5565, 1821.0],
    [0.5575, 1846.0],
    [0.5585, 1787.0],
    [0.5595, 1808.0],
    [0.5605, 1843.0],
    [0.5615, 1824.0],
    [0.5625, 1850.0],
    [0.5635, 1861.0],
    [0.5645, 1854.0],
    [0.5655, 1798.0],
    [0.5665, 1829.0],
    [0.5675, 1887.0],
    [0.5685, 1810.0],
    [0.5695, 1860.0],
    [0.5705, 1769.0],
    [0.5715, 1823.0],
    [0.5725, 1892.0],
    [0.5735, 1876.0],
    [0.5745, 1867.0],
    [0.5755, 1830.0],
    [0.5765, 1846.0],
    [0.5775, 1857.0],
    [0.5785, 1783.0],
    [0.5795, 1828.0],
    [0.5805, 1838.0],
    [0.5815, 1853.0],
    [0.5825, 1873.0],
    [0.5835, 1857.0],
    [0.5845, 1860.0],
    [0.5855, 1783.0],
    [0.5865, 1830.0],
    [0.5875, 1848.0],
    [0.5885, 1750.0],
    [0.5895, 1612.0],
    [0.5905, 1813.0],
    [0.5915, 1787.0],
    [0.5925, 1808.0],
    [0.5935, 1796.0],
    [0.5945, 1773.0],
    [0.5955, 1782.0],
    [0.5965, 1805.0],
    [0.5975, 1780.0],
    [0.5985, 1757.0],
    [0.5995, 1774.0],
    [0.6005, 1746.0],
    [0.6015, 1751.0],
    [0.6025, 1719.0],
    [0.6035, 1787.0],
    [0.6045, 1776.0],
    [0.6055, 1763.0],
    [0.6065, 1759.0],
    [0.6075, 1757.0],
    [0.6085, 1743.0],
    [0.6095, 1744.0],
    [0.6105, 1703.0],
    [0.6115, 1746.0],
    [0.6125, 1705.0],
    [0.6135, 1683.0],
    [0.6145, 1713.0],
    [0.6155, 1713.0],
    [0.6165, 1609.0],
    [0.6175, 1707.0],
    [0.6185, 1724.0],
    [0.6195, 1707.0],
    [0.6205, 1734.0],
    [0.6215, 1690.0],
    [0.6225, 1713.0],
    [0.6235, 1666.0],
    [0.6245, 1656.0],
    [0.6255, 1632.0],
    [0.6265, 1697.0],
    [0.6275, 1697.0],
    [0.6285, 1697.0],
    [0.6295, 1677.0],
    [0.631, 1639.0],
    [0.633, 1651.0],
    [0.635, 1656.0],
    [0.637, 1654.0],
    [0.639, 1651.0],
    [0.641, 1614.0],
    [0.643, 1621.0],
    [0.645, 1627.0],
    [0.647, 1603.0],
    [0.649, 1558.0],
    [0.651, 1606.0],
    [0.653, 1599.0],
    [0.655, 1532.0],
    [0.657, 1384.0],
    [0.659, 1549.0],
    [0.661, 1571.0],
    [0.663, 1555.0],
    [0.665, 1560.0],
    [0.667, 1535.0],
    [0.669, 1546.0],
    [0.671, 1516.0],
    [0.673, 1521.0],
    [0.675, 1510.0],
    [0.677, 1508.0],
    [0.679, 1498.0],
    [0.681, 1492.0],
    [0.683, 1479.0],
    [0.685, 1455.0],
    [0.687, 1467.0],
    [0.689, 1461.0],
    [0.691, 1448.0],
    [0.693, 1448.0],
    [0.695, 1436.0],
    [0.697, 1416.0],
    [0.699, 1425.0],
    [0.701, 1386.0],
    [0.703, 1388.0],
    [0.705, 1415.0],
    [0.707, 1400.0],
    [0.709, 1384.0],
    [0.711, 1385.0],
    [0.713, 1373.0],
    [0.715, 1366.0],
    [0.717, 1354.0],
    [0.719, 1328.0],
    [0.721, 1331.0],
    [0.723, 1348.0],
    [0.725, 1350.0],
    [0.727, 1346.0],
    [0.729, 1319.0],
    [0.731, 1326.0],
    [0.733, 1318.0],
    [0.735, 1309.0],
    [0.737, 1307.0],
    [0.739, 1278.0],
    [0.741, 1258.0],
    [0.743, 1286.0],
    [0.745, 1279.0],
    [0.747, 1283.0],
    [0.749, 1270.0],
    [0.751, 1262.0],
    [0.753, 1259.0],
    [0.755, 1255.0],
    [0.757, 1248.0],
    [0.759, 1240.0],
    [0.761, 1237.0],
    [0.763, 1241.0],
    [0.765, 1221.0],
    [0.767, 1185.0],
    [0.769, 1203.0],
    [0.771, 1204.0],
    [0.773, 1208.0],
    [0.775, 1188.0],
    [0.777, 1196.0],
    [0.779, 1187.0],
    [0.781, 1187.0],
    [0.783, 1176.0],
    [0.785, 1180.0],
    [0.787, 1177.0],
    [0.789, 1174.0],
    [0.791, 1158.0],
    [0.793, 1143.0],
    [0.795, 1134.0],
    [0.797, 1152.0],
    [0.799, 1135.0],
    [0.801, 1142.0],
    [0.803, 1129.0],
    [0.805, 1115.0],
    [0.807, 1120.0],
    [0.809, 1095.0],
    [0.811, 1114.0],
    [0.813, 1115.0],
    [0.815, 1107.0],
    [0.817, 1104.0],
    [0.819, 1063.0],
    [0.821, 1080.0],
    [0.823, 1073.0],
    [0.825, 1075.0],
    [0.826, 1080.0],
    [0.828, 1081.0],
    [0.83, 1063.0],
    [0.832, 1051.0],
    [0.834, 1041.0],
    [0.836, 1052.0],
    [0.838, 1044.0],
    [0.84, 1040.0],
    [0.842, 1036.0],
    [0.844, 1024.0],
    [0.846, 1028.0],
    [0.848, 1023.0],
    [0.85, 966.0],
    [0.852, 996.1],
    [0.854, 878.0],
    [0.856, 975.5],
    [0.858, 1005.0],
    [0.86, 996.9],
    [0.862, 994.9],
    [0.864, 999.3],
    [0.866, 886.2],
    [0.868, 939.5],
    [0.87, 974.7],
    [0.872, 983.3],
    [0.874, 971.3],
    [0.876, 964.0],
    [0.878, 974.9],
    [0.88, 955.4],
    [0.882, 951.1],
    [0.884, 957.9],
    [0.886, 938.3],
    [0.888, 944.3],
    [0.89, 953.0],
    [0.892, 939.4],
    [0.894, 933.2],
    [0.896, 938.7],
    [0.898, 933.9],
    [0.9, 915.8],
    [0.902, 891.6],
    [0.904, 928.5],
    [0.906, 917.6],
    [0.908, 902.5],
    [0.91, 891.6],
    [0.912, 896.7],
    [0.914, 907.1],
    [0.916, 900.4],
    [0.918, 895.1],
    [0.92, 890.8],
    [0.922, 863.0],
    [0.924, 858.5],
    [0.926, 861.2],
    [0.928, 876.9],
    [0.93, 867.7],
    [0.932, 865.1],
    [0.934, 864.1],
    [0.936, 854.7],
    [0.938, 858.0],
    [0.94, 843.8],
    [0.942, 825.0],
    [0.944, 832.4],
    [0.946, 837.5],
    [0.948, 840.7],
    [0.95, 836.9],
    [0.952, 831.7],
    [0.954, 808.0],
    [0.956, 808.2],
    [0.958, 818.8],
    [0.96, 815.1],
    [0.962, 808.9],
    [0.964, 801.3],
    [0.966, 794.7],
    [0.968, 796.9],
    [0.97, 795.9],
    [0.972, 793.6],
    [0.974, 781.5],
    [0.976, 782.5],
    [0.978, 777.9],
    [0.98, 774.6],
    [0.982, 776.4],
    [0.984, 769.8],
    [0.986, 766.1],
    [0.988, 761.5],
    [0.99, 754.1],
    [0.992, 756.7],
    [0.994, 755.6],
    [0.996, 752.5],
    [0.998, 751.0],
    [1.0, 747.9],
    [1.002, 746.9],
    [1.004, 726.1],
    [1.006, 713.6],
    [1.008, 733.5],
    [1.01, 731.3],
    [1.012, 726.2],
    [1.014, 721.0],
    [1.016, 713.9],
    [1.018, 710.7],
    [1.02, 704.1],
    [1.022, 702.1],
    [1.024, 705.4],
    [1.026, 702.7],
    [1.028, 698.9],
    [1.03, 693.7],
    [1.032, 690.5],
    [1.034, 681.7],
    [1.036, 684.0],
    [1.038, 677.2],
    [1.04, 676.1],
    [1.042, 674.6],
    [1.044, 671.4],
    [1.046, 660.0],
    [1.048, 664.4],
    [1.05, 662.2],
    [1.052, 658.6],
    [1.054, 654.9],
    [1.056, 655.7],
    [1.058, 645.1],
    [1.06, 641.5],
    [1.062, 643.8],
    [1.064, 645.9],
    [1.066, 639.5],
    [1.068, 631.7],
    [1.07, 624.1],
    [1.072, 632.6],
    [1.074, 627.6],
    [1.076, 628.0],
    [1.078, 627.2],
    [1.08, 624.7],
    [1.082, 609.9],
    [1.084, 618.0],
    [1.086, 620.8],
    [1.088, 610.3],
    [1.09, 619.9],
    [1.092, 615.9],
    [1.094, 584.9],
    [1.096, 598.3],
    [1.098, 596.1],
    [1.1, 604.2],
    [1.102, 593.2],
    [1.104, 597.4],
    [1.106, 594.5],
    [1.108, 591.6],
    [1.11, 590.6],
    [1.112, 584.3],
    [1.114, 584.4],
    [1.116, 583.1],
    [1.118, 581.5],
    [1.12, 574.1],
    [1.122, 579.6],
    [1.124, 576.9],
    [1.126, 565.5],
    [1.128, 570.0],
    [1.13, 565.3],
    [1.132, 567.8],
    [1.134, 563.8],
    [1.136, 565.8],
    [1.138, 556.9],
    [1.14, 553.0],
    [1.142, 553.1],
    [1.144, 551.4],
    [1.146, 554.8],
    [1.148, 552.5],
    [1.15, 548.9],
    [1.152, 545.8],
    [1.154, 547.9],
    [1.156, 545.5],
    [1.158, 543.5],
    [1.16, 532.0],
    [1.162, 532.5],
    [1.164, 533.2],
    [1.166, 530.3],
    [1.168, 531.2],
    [1.17, 527.6],
    [1.172, 531.5],
    [1.174, 527.3],
    [1.176, 518.4],
    [1.178, 519.0],
    [1.18, 523.9],
    [1.182, 515.9],
    [1.184, 510.3],
    [1.186, 518.7],
    [1.188, 507.5],
    [1.19, 508.5],
    [1.192, 516.1],
    [1.194, 514.5],
    [1.196, 508.4],
    [1.198, 494.3],
    [1.2, 500.3],
    [1.202, 506.8],
    [1.204, 494.8],
    [1.206, 503.9],
    [1.208, 489.0],
    [1.21, 488.2],
    [1.212, 493.3],
    [1.214, 494.2],
    [1.216, 493.0],
    [1.218, 489.7],
    [1.22, 487.5],
    [1.222, 485.4],
    [1.224, 484.6],
    [1.226, 481.7],
    [1.228, 477.1],
    [1.23, 479.2],
    [1.232, 475.0],
    [1.234, 472.9],
    [1.236, 471.9],
    [1.238, 470.3],
    [1.24, 465.3],
    [1.242, 464.2],
    [1.244, 461.9],
    [1.246, 463.5],
    [1.248, 463.3],
    [1.25, 462.4],
    [1.252, 457.1],
    [1.254, 457.4],
    [1.256, 455.1],
    [1.258, 453.3],
    [1.26, 453.0],
    [1.262, 449.7],
    [1.264, 447.8],
    [1.266, 446.7],
    [1.268, 441.7],
    [1.27, 445.3],
    [1.272, 445.2],
    [1.274, 443.1],
    [1.276, 445.1],
    [1.278, 444.0],
    [1.28, 435.6],
    [1.282, 401.4],
    [1.284, 425.9],
    [1.286, 432.8],
    [1.288, 431.4],
    [1.29, 425.5],
    [1.292, 425.4],
    [1.294, 422.3],
    [1.296, 422.4],
    [1.298, 418.4],
    [1.3, 418.6],
    [1.302, 413.9],
    [1.304, 411.1],
    [1.306, 413.6],
    [1.308, 412.3],
    [1.31, 410.6],
    [1.312, 403.3],
    [1.314, 402.2],
    [1.316, 397.9],
    [1.318, 401.7],
    [1.32, 401.6],
    [1.322, 398.6],
    [1.324, 398.1],
    [1.326, 394.9],
    [1.328, 390.8],
    [1.33, 387.8],
    [1.332, 386.3],
    [1.334, 389.2],
    [1.336, 386.6],
    [1.338, 383.2],
    [1.34, 379.0],
    [1.342, 380.5],
    [1.344, 379.8],
    [1.346, 377.2],
    [1.348, 376.6],
    [1.35, 372.4],
    [1.352, 374.2],
    [1.354, 372.2],
    [1.356, 367.5],
    [1.358, 368.8],
    [1.36, 367.3],
    [1.362, 367.7],
    [1.364, 365.7],
    [1.366, 365.7],
    [1.368, 362.8],
    [1.37, 359.9],
    [1.372, 362.1],
    [1.374, 361.1],
    [1.376, 356.1],
    [1.378, 358.0],
    [1.38, 357.9],
    [1.382, 354.5],
    [1.384, 354.7],
    [1.386, 353.2],
    [1.388, 353.0],
    [1.39, 350.6],
    [1.392, 351.3],
    [1.394, 348.8],
    [1.396, 348.7],
    [1.398, 349.2],
    [1.4, 342.7],
    [1.402, 343.9],
    [1.404, 342.8],
    [1.406, 343.1],
    [1.408, 342.7],
    [1.41, 341.8],
    [1.412, 334.8],
    [1.414, 337.7],
    [1.416, 338.5],
    [1.418, 338.6],
    [1.42, 335.7],
    [1.422, 331.5],
    [1.424, 331.1],
    [1.426, 328.1],
    [1.428, 328.5],
    [1.43, 325.7],
    [1.432, 330.0],
    [1.434, 328.4],
    [1.436, 328.5],
    [1.438, 328.3],
    [1.44, 318.8],
    [1.442, 318.6],
    [1.444, 319.7],
    [1.446, 321.6],
    [1.448, 321.6],
    [1.45, 318.7],
    [1.452, 315.4],
    [1.454, 314.3],
    [1.456, 313.1],
    [1.458, 316.7],
    [1.46, 315.6],
    [1.462, 312.1],
    [1.464, 310.5],
    [1.466, 310.8],
    [1.468, 311.4],
    [1.47, 310.2],
    [1.472, 307.3],
    [1.474, 303.4],
    [1.476, 304.8],
    [1.478, 304.4],
    [1.48, 306.8],
    [1.482, 304.4],
    [1.484, 303.9],
    [1.486, 303.3],
    [1.488, 285.5],
    [1.49, 301.5],
    [1.492, 301.8],
    [1.494, 303.3],
    [1.496, 297.2],
    [1.498, 299.4],
    [1.5, 301.1],
    [1.502, 292.4],
    [1.504, 279.9],
    [1.506, 284.8],
    [1.508, 291.9],
    [1.51, 294.7],
    [1.512, 291.3],
    [1.514, 288.3],
    [1.516, 288.2],
    [1.518, 288.4],
    [1.52, 286.6],
    [1.522, 282.4],
    [1.524, 283.5],
    [1.526, 284.6],
    [1.528, 284.6],
    [1.53, 276.5],
    [1.532, 282.3],
    [1.534, 278.4],
    [1.536, 280.6],
    [1.538, 277.3],
    [1.54, 273.0],
    [1.542, 275.3],
    [1.544, 277.8],
    [1.546, 277.2],
    [1.548, 271.1],
    [1.55, 271.3],
    [1.552, 273.1],
    [1.554, 267.6],
    [1.556, 267.1],
    [1.558, 268.9],
    [1.56, 268.3],
    [1.562, 269.7],
    [1.564, 266.9],
    [1.566, 265.4],
    [1.568, 263.3],
    [1.57, 264.5],
    [1.572, 267.3],
    [1.574, 261.0],
    [1.576, 253.6],
    [1.578, 254.7],
    [1.58, 265.0],
    [1.582, 259.0],
    [1.584, 259.1],
    [1.586, 259.9],
    [1.588, 249.0],
    [1.59, 240.5],
    [1.592, 252.6],
    [1.594, 258.3],
    [1.596, 250.6],
    [1.598, 254.5],
    [1.6, 251.2],
    [1.602, 248.9],
    [1.604, 249.7],
    [1.606, 247.7],
    [1.608, 249.1],
    [1.61, 240.0],
    [1.612, 243.0],
    [1.614, 244.9],
    [1.616, 237.4],
    [1.618, 242.3],
    [1.62, 236.9],
    [1.622, 238.3],
    [1.624, 241.6],
    [1.626, 240.2],
    [1.628, 241.8],
    [1.63, 239.3],
    [1.632, 238.7],
    [1.634, 235.9],
    [1.636, 235.7],
    [1.638, 227.4],
    [1.64, 226.2],
    [1.642, 226.6],
    [1.644, 227.8],
    [1.646, 229.4],
    [1.648, 229.2],
    [1.65, 227.2],
    [1.652, 226.8],
    [1.654, 226.2],
    [1.656, 226.0],
    [1.658, 225.2],
    [1.66, 224.5],
    [1.662, 224.6],
    [1.664, 222.7],
    [1.666, 221.2],
    [1.668, 219.3],
    [1.67, 222.5],
    [1.672, 217.3],
    [1.674, 219.3],
    [1.676, 216.1],
    [1.678, 216.8],
    [1.68, 208.0],
    [1.682, 205.4],
    [1.684, 212.9],
    [1.686, 213.1],
    [1.688, 212.0],
    [1.69, 210.5],
    [1.692, 212.3],
    [1.694, 211.2],
    [1.696, 210.0],
    [1.698, 208.9],
    [1.7, 206.3],
    [1.702, 204.7],
    [1.704, 205.2],
    [1.706, 205.0],
    [1.708, 201.7],
    [1.71, 201.3],
    [1.712, 198.2],
    [1.714, 203.7],
    [1.716, 202.2],
    [1.718, 201.0],
    [1.72, 199.3],
    [1.722, 197.5],
    [1.724, 195.4],
    [1.726, 198.2],
    [1.728, 197.1],
    [1.73, 198.4],
    [1.732, 193.6],
    [1.734, 187.4],
    [1.736, 182.7],
    [1.738, 186.3],
    [1.74, 190.5],
    [1.742, 190.2],
    [1.744, 190.7],
    [1.746, 186.7],
    [1.748, 187.2],
    [1.75, 185.8],
    [1.752, 185.0],
    [1.754, 185.6],
    [1.756, 184.9],
    [1.758, 184.3],
    [1.76, 183.1],
    [1.762, 179.3],
    [1.764, 180.7],
    [1.766, 181.7],
    [1.768, 180.2],
    [1.77, 179.1],
    [1.772, 179.4],
    [1.774, 179.2],
    [1.776, 176.3],
    [1.778, 174.7],
    [1.78, 175.6],
    [1.782, 174.7],
    [1.784, 173.5],
    [1.786, 173.9],
    [1.788, 174.7],
    [1.79, 173.3],
    [1.792, 172.1],
    [1.794, 170.9],
    [1.796, 170.6],
    [1.798, 170.3],
    [1.8, 169.9],
    [1.802, 167.2],
    [1.804, 168.8],
    [1.806, 168.8],
    [1.808, 168.5],
    [1.81, 168.6],
    [1.812, 167.5],
    [1.814, 165.8],
    [1.816, 160.5],
    [1.818, 152.0],
    [1.82, 159.6],
    [1.822, 159.8],
    [1.824, 162.4],
    [1.826, 162.8],
    [1.828, 161.1],
    [1.83, 160.6],
    [1.832, 159.3],
    [1.834, 158.5],
    [1.836, 158.1],
    [1.838, 156.2],
    [1.84, 156.2],
    [1.842, 154.0],
    [1.844, 154.1],
    [1.846, 153.5],
    [1.848, 151.0],
    [1.85, 154.6],
    [1.852, 153.4],
    [1.854, 152.5],
    [1.856, 150.9],
    [1.858, 152.5],
    [1.86, 150.3],
    [1.862, 150.4],
    [1.864, 150.9],
    [1.866, 149.4],
    [1.868, 149.2],
    [1.87, 150.8],
    [1.872, 147.3],
    [1.874, 140.1],
    [1.876, 129.9],
    [1.878, 144.1],
    [1.88, 146.2],
    [1.882, 147.4],
    [1.884, 146.4],
    [1.886, 143.9],
    [1.888, 145.3],
    [1.89, 142.4],
    [1.892, 140.8],
    [1.894, 139.6],
    [1.896, 137.3],
    [1.898, 139.0],
    [1.9, 139.7],
    [1.902, 140.9],
    [1.904, 138.6],
    [1.906, 139.0],
    [1.908, 137.7],
    [1.91, 137.8],
    [1.912, 135.4],
    [1.914, 137.0],
    [1.916, 136.0],
    [1.918, 135.3],
    [1.92, 133.3],
    [1.922, 135.0],
    [1.924, 134.1],
    [1.926, 134.4],
    [1.928, 132.2],
    [1.93, 131.3],
    [1.932, 130.8],
    [1.934, 132.0],
    [1.936, 132.8],
    [1.938, 132.1],
    [1.94, 129.9],
    [1.942, 129.4],
    [1.944, 120.3],
    [1.946, 119.2],
    [1.948, 127.1],
    [1.95, 126.1],
    [1.952, 125.5],
    [1.954, 128.6],
    [1.956, 127.6],
    [1.958, 127.1],
    [1.96, 126.1],
    [1.962, 124.0],
    [1.964, 122.2],
    [1.966, 123.1],
    [1.968, 124.0],
    [1.97, 123.9],
    [1.972, 121.3],
    [1.974, 120.8],
    [1.976, 122.4],
    [1.978, 119.4],
    [1.98, 119.6],
    [1.982, 120.5],
    [1.984, 119.7],
    [1.986, 117.8],
    [1.988, 119.5],
    [1.99, 119.8],
    [1.992, 118.0],
    [1.994, 116.2],
    [1.996, 117.3],
    [1.998, 115.9],
    [2.0, 117.0],
    [2.002, 116.1],
    [2.004, 114.8],
    [2.006, 114.7],
    [2.008, 115.4],
    [2.01, 114.9],
    [2.012, 114.5],
    [2.014, 113.8],
    [2.016, 113.7],
    [2.018, 113.4],
    [2.02, 111.6],
    [2.022, 110.7],
    [2.024, 111.6],
    [2.026, 111.5],
    [2.028, 110.7],
    [2.03, 108.6],
    [2.032, 109.8],
    [2.034, 109.2],
    [2.036, 108.3],
    [2.038, 106.4],
    [2.04, 107.8],
    [2.042, 107.6],
    [2.044, 107.6],
    [2.046, 107.1],
    [2.048, 106.3],
    [2.05, 105.9],
    [2.052, 104.7],
    [2.054, 104.6],
    [2.056, 104.6],
    [2.058, 104.0],
    [2.06, 102.8],
    [2.062, 102.3],
    [2.064, 100.5],
    [2.066, 102.5],
    [2.068, 101.9],
    [2.07, 100.3],
    [2.072, 100.4],
    [2.074, 100.9],
    [2.076, 100.6],
    [2.078, 100.0],
    [2.08, 98.78],
    [2.082, 98.64],
    [2.084, 97.72],
    [2.086, 98.52],
    [2.088, 98.35],
    [2.09, 97.88],
    [2.092, 95.67],
    [2.094, 95.93],
    [2.096, 95.8],
    [2.098, 96.2],
    [2.1, 96.06],
    [2.102, 95.77],
    [2.104, 95.59],
    [2.106, 95.74],
    [2.108, 95.13],
    [2.11, 93.96],
    [2.112, 94.52],
    [2.114, 94.36],
    [2.116, 93.31],
    [2.118, 93.11],
    [2.12, 92.75],
    [2.122, 92.75],
    [2.124, 91.89],
    [2.126, 92.08],
    [2.128, 92.25],
    [2.13, 92.09],
    [2.132, 92.1],
    [2.134, 91.55],
    [2.136, 90.12],
    [2.138, 91.1],
    [2.14, 90.83],
    [2.142, 90.64],
    [2.144, 90.06],
    [2.146, 89.39],
    [2.148, 89.79],
    [2.15, 89.57],
    [2.152, 89.13],
    [2.154, 88.78],
    [2.156, 88.74],
    [2.158, 88.42],
    [2.16, 87.81],
    [2.162, 86.86],
    [2.164, 84.56],
    [2.166, 78.49],
    [2.168, 83.0],
    [2.17, 85.57],
    [2.172, 85.91],
    [2.174, 85.92],
    [2.176, 85.32],
    [2.178, 84.25],
    [2.18, 84.97],
    [2.182, 84.25],
    [2.184, 84.57],
    [2.186, 84.65],
    [2.188, 82.77],
    [2.19, 83.04],
    [2.192, 83.77],
    [2.194, 83.49],
    [2.196, 83.18],
    [2.198, 82.99],
    [2.2, 82.65],
    [2.202, 82.3],
    [2.204, 82.11],
    [2.206, 79.66],
    [2.208, 79.66],
    [2.21, 80.8],
    [2.212, 81.05],
    [2.214, 80.72],
    [2.216, 79.94],
    [2.218, 79.7],
    [2.22, 79.97],
    [2.222, 79.62],
    [2.224, 79.26],
    [2.226, 78.11],
    [2.228, 78.26],
    [2.23, 78.31],
    [2.232, 78.15],
    [2.234, 78.02],
    [2.236, 77.58],
    [2.238, 76.48],
    [2.24, 76.39],
    [2.242, 76.42],
    [2.244, 76.24],
    [2.246, 76.12],
    [2.248, 75.2],
    [2.25, 75.41],
    [2.252, 75.12],
    [2.254, 74.02],
    [2.256, 74.22],
    [2.258, 74.41],
    [2.26, 74.21],
    [2.262, 72.99],
    [2.264, 73.29],
    [2.266, 73.15],
    [2.268, 73.27],
    [2.27, 72.97],
    [2.272, 72.77],
    [2.274, 72.52],
    [2.276, 72.39],
    [2.278, 72.42],
    [2.28, 71.65],
    [2.282, 70.07],
    [2.284, 71.25],
    [2.286, 71.24],
    [2.288, 71.27],
    [2.29, 71.1],
    [2.292, 70.67],
    [2.294, 69.2],
    [2.296, 69.08],
    [2.298, 69.19],
    [2.3, 69.53],
    [2.302, 69.55],
    [2.304, 69.31],
    [2.306, 69.23],
    [2.308, 69.01],
    [2.31, 68.7],
    [2.312, 68.67],
    [2.314, 68.26],
    [2.316, 67.79],
    [2.318, 67.45],
    [2.32, 67.68],
    [2.322, 66.75],
    [2.324, 65.36],
    [2.326, 65.59],
    [2.328, 66.29],
    [2.33, 66.16],
    [2.332, 65.84],
    [2.334, 65.71],
    [2.336, 65.36],
    [2.338, 64.96],
    [2.34, 65.2],
    [2.342, 65.39],
    [2.344, 65.09],
    [2.346, 64.86],
    [2.348, 64.72],
    [2.35, 64.53],
    [2.352, 62.89],
    [2.354, 62.39],
    [2.356, 62.82],
    [2.358, 62.66],
    [2.36, 63.08],
    [2.362, 63.05],
    [2.364, 62.95],
    [2.366, 62.84],
    [2.368, 62.63],
    [2.37, 62.11],
    [2.372, 62.07],
    [2.374, 60.66],
    [2.376, 61.64],
    [2.378, 61.92],
    [2.38, 61.72],
    [2.382, 60.98],
    [2.384, 58.85],
    [2.386, 59.08],
    [2.388, 60.04],
    [2.39, 60.29],
    [2.392, 60.08],
    [2.394, 60.03],
    [2.396, 59.96],
    [2.398, 59.89],
    [2.4, 59.44],
    [2.402, 59.65],
    [2.404, 59.45],
    [2.406, 59.19],
    [2.408, 59.15],
    [2.41, 59.02],
    [2.412, 58.94],
    [2.414, 57.34],
    [2.416, 55.99],
    [2.418, 57.48],
    [2.42, 57.7],
    [2.422, 57.67],
    [2.424, 57.26],
    [2.426, 57.17],
    [2.428, 57.12],
    [2.43, 57.12],
    [2.432, 57.02],
    [2.434, 56.41],
    [2.436, 56.18],
    [2.438, 55.99],
    [2.44, 56.39],
    [2.442, 56.17],
    [2.444, 56.03],
    [2.446, 54.98],
    [2.448, 54.57],
    [2.45, 54.62],
    [2.452, 54.32],
    [2.454, 54.55],
    [2.456, 53.7],
    [2.458, 53.92],
    [2.46, 54.57],
    [2.462, 54.42],
    [2.464, 54.35],
    [2.466, 54.05],
    [2.468, 53.9],
    [2.47, 52.85],
    [2.472, 53.3],
    [2.474, 53.13],
    [2.476, 53.43],
    [2.478, 53.03],
    [2.48, 51.77],
    [2.482, 51.4],
    [2.484, 52.19],
    [2.486, 51.6],
    [2.488, 51.69],
    [2.49, 52.25],
    [2.492, 51.98],
    [2.494, 51.75],
    [2.496, 51.52],
    [2.498, 51.54],
    [2.5, 51.55],
    [2.52, 49.84],
    [2.54, 48.14],
    [2.56, 46.72],
    [2.58, 45.5],
    [2.6, 44.57],
    [2.62, 43.05],
    [2.64, 42.11],
    [2.66, 40.79],
    [2.68, 39.68],
    [2.7, 38.67],
    [2.72, 37.63],
    [2.74, 36.63],
    [2.76, 35.46],
    [2.78, 34.68],
    [2.8, 33.85],
    [2.82, 32.97],
    [2.84, 32.09],
    [2.86, 31.19],
    [2.88, 30.32],
    [2.9, 29.69],
    [2.92, 28.9],
    [2.94, 28.17],
    [2.96, 27.5],
    [2.98, 26.82],
    [3.0, 26.12],
    [3.02, 25.47],
    [3.04, 24.65],
    [3.06, 24.22],
    [3.08, 23.64],
    [3.1, 23.06],
    [3.12, 22.46],
    [3.140, 21.98],
    [3.16, 21.44],
    [3.18, 20.96],
    [3.2, 20.48],
    [3.22, 20.0],
    [3.24, 19.51],
    [3.26, 19.07],
    [3.28, 18.58],
    [3.3, 18.02],
    [3.32, 17.68],
    [3.34, 17.37],
    [3.36, 16.97],
    [3.38, 16.59],
    [3.4, 16.15],
    [3.42, 15.84],
    [3.44, 15.54],
    [3.46, 15.2],
    [3.48, 14.86],
    [3.5, 14.56],
    [3.52, 14.25],
    [3.54, 13.93],
    [3.56, 13.62],
    [3.58, 13.34],
    [3.6, 13.07],
    [3.62, 12.81],
    [3.64, 12.51],
    [3.66, 12.22],
    [3.68, 11.93],
    [3.7, 11.62],
    [3.72, 11.45],
    [3.74, 11.08],
    [3.76, 10.96],
    [3.78, 10.78],
    [3.8, 10.57],
    [3.82, 10.38],
    [3.84, 10.19],
    [3.86, 9.983],
    [3.88, 9.782],
    [3.9, 9.599],
    [3.92, 9.427],
    [3.94, 9.233],
    [3.96, 9.032],
    [3.98, 8.857],
    [4.0, 8.669],
    [4.02, 8.557],
    [4.04, 8.385],
    [4.06, 8.217],
    [4.08, 8.054],
    [4.1, 7.894],
    [4.12, 7.739],
    [4.14, 7.587],
    [4.16, 7.439],
    [4.18, 7.294],
    [4.2, 7.153],
    [4.22, 7.015],
    [4.24, 6.881],
    [4.26, 6.749],
    [4.28, 6.621],
    [4.3, 6.496],
    [4.32, 6.374],
    [4.34, 6.254],
    [4.36, 6.138],
    [4.38, 6.024],
    [4.4, 5.913],
    [4.42, 5.804],
    [4.44, 5.698],
    [4.46, 5.594],
    [4.48, 5.492],
    [4.5, 5.393],
    [4.52, 5.296],
    [4.54, 5.201],
    [4.56, 5.108],
    [4.58, 5.018],
    [4.6, 4.929],
    [4.62, 4.842],
    [4.64, 4.757],
    [4.66, 4.674],
    [4.68, 4.593],
    [4.7, 4.514],
    [4.72, 4.436],
    [4.74, 4.36],
    [4.76, 4.285],
    [4.78, 4.212],
    [4.8, 4.141],
    [4.82, 4.071],
    [4.84, 4.003],
    [4.86, 3.936],
    [4.88, 3.87],
    [4.9, 3.806],
    [4.92, 3.743],
    [4.94, 3.681],
    [4.96, 3.621],
    [4.98, 3.562],
    [5.0, 3.504],
    [5.05, 3.394],
    [5.1, 3.267],
    [5.15, 3.146],
    [5.2, 3.03],
    [5.25, 2.92],
    [5.3, 2.815],
    [5.35, 2.715],
    [5.4, 2.619],
    [5.45, 2.527],
    [5.5, 2.439],
    [5.55, 2.355],
    [5.6, 2.275],
    [5.65, 2.198],
    [5.7, 2.124],
    [5.75, 2.054],
    [5.8, 1.986],
    [5.85, 1.921],
    [5.9, 1.859],
    [5.95, 1.799],
    [6.0, 1.742],
    [6.05, 1.687],
    [6.1, 1.634],
    [6.15, 1.583],
    [6.2, 1.534],
    [6.25, 1.487],
    [6.3, 1.442],
    [6.35, 1.399],
    [6.4, 1.357],
    [6.45, 1.317],
    [6.5, 1.278],
    [6.55, 1.24],
    [6.6, 1.204],
    [6.65, 1.17],
    [6.7, 1.136],
    [6.75, 1.104],
    [6.8, 1.073],
    [6.85, 1.043],
    [6.9, 1.014],
    [6.95, 0.9862],
    [7.0, 0.9592],
    [7.05, 0.9331],
    [7.1, 0.908],
    [7.15, 0.8836],
    [7.2, 0.8601],
    [7.25, 0.8374],
    [7.3, 0.8154],
    [7.35, 0.7942],
    [7.4, 0.7736],
    [7.45, 0.7537],
    [7.5, 0.7344],
    [7.55, 0.7158],
    [7.6, 0.6977],
    [7.65, 0.6802],
    [7.7, 0.6633],
    [7.75, 0.6469],
    [7.8, 0.631],
    [7.85, 0.6156],
    [7.9, 0.6006],
    [7.95, 0.5862],
    [8.0, 0.5721],
    [8.05, 0.5585],
    [8.1, 0.5453],
    [8.15, 0.5324],
    [8.2, 0.52],
    [8.25, 0.5079],
    [8.3, 0.4961],
    [8.35, 0.4847],
    [8.4, 0.4737],
    [8.45, 0.4629],
    [8.5, 0.4525],
    [8.55, 0.4423],
    [8.6, 0.4324],
    [8.65, 0.4228],
    [8.7, 0.4135],
    [8.75, 0.4044],
    [8.8, 0.3956],
    [8.85, 0.387],
    [8.9, 0.3787],
    [8.95, 0.3706],
    [9.0, 0.3627],
    [9.05, 0.355],
    [9.1, 0.3475],
    [9.15, 0.3402],
    [9.2, 0.3331],
    [9.25, 0.3262],
    [9.3, 0.3195],
    [9.35, 0.3129],
    [9.4, 0.3065],
    [9.45, 0.3003],
    [9.5, 0.2942],
    [9.55, 0.2883],
    [9.6, 0.2825],
    [9.65, 0.2769],
    [9.7, 0.2714],
    [9.75, 0.2661],
    [9.8, 0.2608],
    [9.85, 0.2558],
    [9.9, 0.2508],
    [9.95, 0.246],
    [10., 0.2412],
    [11., 0.1635],
    [12., 0.1152],
    [13., 0.0834],
    [14., 0.0619],
    [15., 0.0469],
    [16., 0.0361],
    [17., 0.0283],
    [18., 0.0225],
    [19., 0.0181],
    [20., 0.0147],
    [25., 0.00605],
    [30., 0.00293],
    [35., 0.00158],
    [40., 0.000931],
    [50., 0.000383],
    [60., 0.000185],
    [80., 5.88e-05],
    [100., 2.42e-05],
    [120., 1.17e-05],
    [150., 4.81e-06],
    [200., 1.53e-06],
    [250., 6.28e-07],
    [300., 2.95e-07],
    [400., 1.01e-07],
    [1000., 3.38e-09],
];
